#include <iostream>

#include "snake.h"

#include <mapbox/polylabel.hpp>
#include <mapbox/geometry.hpp>

#include <boost/geometry.hpp>
#include <boost/geometry/geometries/polygon.hpp>
#include <boost/geometry/geometries/adapted/boost_tuple.hpp>

using namespace operations_research;

#ifndef NDEBUG
//#define SHOW_TIME
#endif

namespace bg = boost::geometry;
namespace trans = bg::strategy::transform;

namespace snake {
//=========================================================================
// Geometry stuff.
//=========================================================================

BOOST_GEOMETRY_REGISTER_BOOST_TUPLE_CS(cs::cartesian)

void polygonCenter(const BoostPolygon &polygon, BoostPoint &center)
{
 using namespace mapbox;
if (polygon.outer().empty())
    return;
geometry::polygon<double> p;
geometry::linear_ring<double> lr1;
for (size_t i = 0; i < polygon.outer().size(); ++i) {
    geometry::point<double> vertex(polygon.outer()[i].get<0>(), polygon.outer()[i].get<1>());
    lr1.push_back(vertex);
}
p.push_back(lr1);
geometry::point<double> c = polylabel(p);

center.set<0>(c.x);
center.set<1>(c.y);
}

void minimalBoundingBox(const BoostPolygon &polygon, BoundingBox &minBBox)
{
 /*
 Find the minimum-area bounding box of a set of 2D points

 The input is a 2D convex hull, in an Nx2 numpy array of x-y co-ordinates.
 The first and last points points must be the same, making a closed polygon.
 This program finds the rotation angles of each edge of the convex polygon,
 then tests the area of a bounding box aligned with the unique angles in
 90 degrees of the 1st Quadrant.
 Returns the

 Tested with Python 2.6.5 on Ubuntu 10.04.4 (original version)
 Results verified using Matlab

 Copyright (c) 2013, David Butterworth, University of Queensland
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the Willow Garage, Inc. nor the names of its
       contributors may be used to endorse or promote products derived from
       this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
 */

 if (polygon.outer().empty())
     return;
 BoostPolygon convex_hull;
 bg::convex_hull(polygon, convex_hull);

 //cout << "Convex hull: " << bg::wkt<BoostPolygon2D>(convex_hull) << endl;

 //# Compute edges (x2-x1,y2-y1)
 std::vector<BoostPoint> edges;
 auto convex_hull_outer = convex_hull.outer();
 for (long i=0; i < long(convex_hull_outer.size())-1; ++i) {
     BoostPoint p1      = convex_hull_outer.at(i);
     BoostPoint p2      = convex_hull_outer.at(i+1);
     double edge_x   = p2.get<0>() - p1.get<0>();
     double edge_y   = p2.get<1>() - p1.get<1>();
     edges.push_back(BoostPoint{edge_x, edge_y});
 }

//    cout << "Edges: ";
//    for (auto e : edges)
//        cout << e.get<0>() << " " << e.get<1>() << ",";
//    cout << endl;

 // Calculate unique edge angles  atan2(y/x)
 double angle_scale = 1e3;
 std::set<long> angles_long;
 for (auto vertex : edges) {
     double angle = std::fmod(atan2(vertex.get<1>(), vertex.get<0>()), M_PI / 2);
     angle = angle < 0 ? angle + M_PI / 2 : angle; // want strictly positive answers
     angles_long.insert(long(round(angle*angle_scale)));
 }
 std::vector<double> edge_angles;
 for (auto a : angles_long)
     edge_angles.push_back(double(a)/angle_scale);


//    cout << "Unique angles: ";
//    for (auto e : edge_angles)
//        cout << e*180/M_PI << ",";
//    cout << endl;

 double min_area = std::numeric_limits<double>::infinity();
 // Test each angle to find bounding box with smallest area
 // print "Testing", len(edge_angles), "possible rotations for bounding box... \n"
 for (double angle : edge_angles){

     trans::rotate_transformer<bg::degree, double, 2, 2> rotate(angle*180/M_PI);
     BoostPolygon hull_rotated;
     bg::transform(convex_hull, hull_rotated, rotate);
     //cout << "Convex hull rotated: " << bg::wkt<BoostPolygon2D>(hull_rotated) << endl;

     bg::model::box<BoostPoint> box;
     bg::envelope(hull_rotated, box);
//        cout << "Bounding box: " << bg::wkt<bg::model::box<BoostPoint2D>>(box) << endl;

     //# print "Rotated hull points are \n", rot_points
     BoostPoint min_corner = box.min_corner();
     BoostPoint max_corner = box.max_corner();
     double min_x = min_corner.get<0>();
     double max_x = max_corner.get<0>();
     double min_y = min_corner.get<1>();
     double max_y = max_corner.get<1>();
//        cout << "min_x: " << min_x << endl;
//        cout << "max_x: " << max_x << endl;
//        cout << "min_y: " << min_y << endl;
//        cout << "max_y: " << max_y << endl;

     // Calculate height/width/area of this bounding rectangle
     double width    = max_x - min_x;
     double height   = max_y - min_y;
     double area     = width * height;
//        cout << "Width: " << width << endl;
//        cout << "Height: " << height << endl;
//        cout << "area: " << area << endl;
//        cout << "angle: " << angle*180/M_PI << endl;

     // Store the smallest rect found first (a simple convex hull might have 2 answers with same area)
     if (area < min_area){
         min_area = area;
         minBBox.angle  = angle;
         minBBox.width  = width;
         minBBox.height = height;

         minBBox.corners.clear();
         minBBox.corners.outer().push_back(BoostPoint{min_x, min_y});
         minBBox.corners.outer().push_back(BoostPoint{min_x, max_y});
         minBBox.corners.outer().push_back(BoostPoint{max_x, max_y});
         minBBox.corners.outer().push_back(BoostPoint{max_x, min_y});
         minBBox.corners.outer().push_back(BoostPoint{min_x, min_y});
     }
     //cout << endl << endl;
 }


 // Transform corners of minimal bounding box.
 trans::rotate_transformer<bg::degree, double, 2, 2> rotate(-minBBox.angle*180/M_PI);
 BoostPolygon rotated_polygon;
 bg::transform(minBBox.corners, rotated_polygon, rotate);
 minBBox.corners = rotated_polygon;
}

void toBoost(const Point2D &point, BoostPoint &boost_point)
{
 boost_point.set<0>(point[0]);
 boost_point.set<1>(point[1]);
}

void fromBoost(const BoostPoint &boost_point, Point2D &point)
{
 point[0] = boost_point.get<0>();
 point[1] = boost_point.get<1>();
}

void toBoost(const Point2DList &point_list, BoostPolygon &boost_polygon)
{
 for (auto vertex : point_list) {
     BoostPoint boost_vertex;
     toBoost(vertex, boost_vertex);
     boost_polygon.outer().push_back(boost_vertex);
 }
 bg::correct(boost_polygon);
}

void fromBoost(const BoostPolygon &boost_polygon, Point2DList &point_list)
{
 for (auto boost_vertex : boost_polygon.outer()) {
     Point2D vertex;
     fromBoost(boost_vertex, vertex);
     point_list.push_back(vertex);
 }
}

void rotateDeg(const Point2DList &point_list, Point2DList &rotated_point_list, double degree)
{
 trans::rotate_transformer<bg::degree, double, 2, 2> rotate(degree);
 BoostPolygon boost_polygon;
 toBoost(point_list, boost_polygon);
 BoostPolygon rotated_polygon;
 bg::transform(boost_polygon, rotated_polygon, rotate);
 fromBoost(rotated_polygon, rotated_point_list);
}

void rotateRad(const Point2DList &point_list, Point2DList &rotated_point_list, double rad)
{
 rotateDeg(point_list, rotated_point_list, rad*180/M_PI);
}

bool isClockwise(const Point2DList &point_list)
{
 double orientaion = 0;
 double len = point_list.size();
 for (long i=0; i < len-1; ++i){
     Point2D v1 = point_list[i];
     Point2D v2 = point_list[i+1];
     orientaion += (v2[0]-v1[0])*(v2[1]+v1[1]);
 }
 Point2D v1 = point_list[len-1];
 Point2D v2 = point_list[0];
 orientaion += (v2[0]-v1[0])*(v2[1]+v1[1]);

 return orientaion > 0 ? true : false;
}

void offsetPolygon(const BoostPolygon &polygon, BoostPolygon &polygonOffset, double offset)
{
 bg::strategy::buffer::distance_symmetric<double> distance_strategy(offset);
 bg::strategy::buffer::join_miter join_strategy(3);
 bg::strategy::buffer::end_flat end_strategy;
 bg::strategy::buffer::point_square point_strategy;
 bg::strategy::buffer::side_straight side_strategy;


 bg::model::multi_polygon<BoostPolygon> result;

 bg::buffer(polygon, result, distance_strategy, side_strategy, join_strategy, end_strategy, point_strategy);

 if (result.size() > 0)
     polygonOffset = result[0];

}


void graphFromPolygon(const BoostPolygon &polygon, const BoostLineString &vertices, Matrix<double> &graph)
{
 size_t n = graph.getN();

 for (size_t i=0; i < n; ++i) {
     BoostPoint v1 = vertices[i];
     for (size_t j=i+1; j < n; ++j){
         BoostPoint v2 = vertices[j];
         BoostLineString path{v1, v2};

         double distance = 0;
         if (!bg::within(path, polygon))
             distance = std::numeric_limits<double>::infinity();
         else
             distance = bg::length(path);

         graph.set(i, j, distance);
         graph.set(j, i, distance);
     }
 }

}

bool dijkstraAlgorithm(const size_t numElements,
                    size_t startIndex,
                    size_t endIndex,
                    std::vector<size_t> &elementPath,
                    std::function<double (const size_t, const size_t)> distanceDij)
{
 if (    startIndex >= numElements
      || endIndex >= numElements
      || endIndex == startIndex) {
     return false;
 }
 // Node struct
 // predecessorIndex is the index of the predecessor node (nodeList[predecessorIndex])
 // distance is the distance between the node and the start node
 // node number is stored by the position in nodeList
 struct Node{
     int predecessorIndex = -1;
     double distance = std::numeric_limits<double>::infinity();
 };

 // The list with all Nodes (elements)
 std::vector<Node> nodeList(numElements);
 // This list will be initalized with indices referring to the elements of nodeList.
 // Elements will be successively remove during the execution of the Dijkstra Algorithm.
 std::vector<size_t> workingSet(numElements);

 //append elements to node list
 for (size_t i = 0; i < numElements; ++i) workingSet[i] = i;


 nodeList[startIndex].distance = 0;

 // Dijkstra Algorithm
 // https://de.wikipedia.org/wiki/Dijkstra-Algorithmus
 while (workingSet.size() > 0) {
     // serach Node with minimal distance
     double minDist = std::numeric_limits<double>::infinity();
     int minDistIndex_WS = -1; // WS = workinSet
     for (size_t i = 0; i < workingSet.size(); ++i) {
         const int nodeIndex = workingSet.at(i);
         const double dist = nodeList.at(nodeIndex).distance;
         if (dist < minDist) {
             minDist = dist;
             minDistIndex_WS = i;
         }
     }
     if (minDistIndex_WS == -1)
             return false;

     size_t indexU_NL = workingSet.at(minDistIndex_WS); // NL = nodeList
     workingSet.erase(workingSet.begin()+minDistIndex_WS);
     if (indexU_NL == endIndex) // shortest path found
         break;

     const double distanceU = nodeList.at(indexU_NL).distance;
     //update distance
     for (size_t i = 0; i < workingSet.size(); ++i) {
         int indexV_NL = workingSet[i]; // NL = nodeList
         Node* v = &nodeList[indexV_NL];
         double dist = distanceDij(indexU_NL, indexV_NL);
         // is ther an alternative path which is shorter?
         double alternative = distanceU + dist;
         if (alternative < v->distance)  {
             v->distance         = alternative;
             v->predecessorIndex = indexU_NL;
         }
     }

 }
 // end Djikstra Algorithm


 // reverse assemble path
 int e = endIndex;
 while (1) {
     if (e == -1) {
         if (elementPath[0] == startIndex)// check if starting point was reached
             break;
         return false;
     }
     elementPath.insert(elementPath.begin(), e);

     //Update Node
     e = nodeList[e].predecessorIndex;

 }
 return true;
}

void toDistanceMatrix(Matrix<double> &graph)
{
 size_t n = graph.getN();

 auto distance = [graph](size_t i, size_t j){
     return graph.get(i,j);
 };


 std::vector<size_t> path;
 for (size_t i=0; i < n; ++i) {
     for (size_t j=i+1; j < n; ++j){
         double d = graph.get(i,j);
         if (!std::isinf(d))
             continue;
         path.clear();
         bool ret = dijkstraAlgorithm(n, i, j, path, distance);
         assert(ret);
         (void)ret;
//            cout << "(" << i << "," << j << ") d: " << d << endl;
//            cout << "Path size: " << path.size() << endl;
//            for (auto idx : path)
//                cout << idx << " ";
//            cout << endl;

         d = 0;
         for (long k=0; k < long(path.size())-1; ++k) {
             size_t idx0 = path[k];
             size_t idx1 = path[k+1];
             double d0 = graph.get(idx0, idx1);
             assert(std::isinf(d0) == false);
             d += d0;
         }

         graph.set(i, j, d);
         graph.set(j, i, d);
     }
 }
}

void shortestPathFromGraph(const Matrix<double> &graph, size_t startIndex, size_t endIndex, std::vector<size_t> &pathIdx)
{

 if (!std::isinf(graph.get(startIndex, endIndex))){
     pathIdx.push_back(startIndex);
     pathIdx.push_back(endIndex);
 } else {
     auto distance = [graph](size_t i, size_t j){
         return graph.get(i, j);
     };
     bool ret = dijkstraAlgorithm(graph.getN(), startIndex, endIndex, pathIdx, distance);
     assert(ret);
     (void)ret;
 }

}

//=========================================================================
// Tile calculation.
//=========================================================================
bool calculateTiles(const BoostPolygon &mArea,
                    double tileWidth,
                    double tileHeight,
                    double minTileArea,
                    std::vector<BoostPolygon> &tiles,
                    std::string &errorString)
{
    using namespace snake_geometry;
    if (tileWidth <= 0 || tileHeight <= 0 || minTileArea < 0) {
        errorString = "Parameters tileWidth, tileHeight, minTileArea must be positive.";
        return false;
    }

    BoundingBox bbox;
    minimalBoundingBox(mArea, bbox);
    double bbox_width   = bbox.width;
    double bbox_height  = bbox.height;
    BoostPoint origin   = bbox.corners.outer()[0];

    //cout << "Origin: " << origin[0] << " " << origin[1] << endl;
    // Transform _measurementAreaENU polygon to bounding box coordinate system.
    trans::rotate_transformer<boost::geometry::degree, double, 2, 2> rotate(bbox.angle*180/M_PI);
    trans::translate_transformer<double, 2, 2> translate(-origin.get<0>(), -origin.get<1>());
    BoostPolygon translated_polygon;
    BoostPolygon rotated_polygon;
    boost::geometry::transform(_measurementAreaENU, translated_polygon, translate);
    boost::geometry::transform(translated_polygon, rotated_polygon, rotate);
    bg::correct(rotated_polygon);

    //cout << bg::wkt<BoostPolygon2D>(rotated_polygon) << endl;

    size_t i_max = ceil(bbox_width/tileWidth);
    size_t j_max = ceil(bbox_height/tileHeight);

    if (i_max < 1 || j_max < 1) {
        errorString = "tileWidth or tileHeight to large.";
        return false;
    }

    trans::rotate_transformer<boost::geometry::degree, double, 2, 2> rotate_back(-bbox.angle*180/M_PI);
    trans::translate_transformer<double, 2, 2> translate_back(origin.get<0>(), origin.get<1>());
    for (size_t i = 0; i < i_max; ++i){
        double x_min = tileWidth*i;
        double x_max = x_min + tileWidth;
        for (size_t j = 0; j < j_max; ++j){
            double y_min = tileHeight*j;
            double y_max = y_min + tileHeight;

            BoostPolygon tile_unclipped;
            tile_unclipped.outer().push_back(BoostPoint{x_min, y_min});
            tile_unclipped.outer().push_back(BoostPoint{x_min, y_max});
            tile_unclipped.outer().push_back(BoostPoint{x_max, y_max});
            tile_unclipped.outer().push_back(BoostPoint{x_max, y_min});
            tile_unclipped.outer().push_back(BoostPoint{x_min, y_min});


            std::deque<BoostPolygon> boost_tiles;
            if (!boost::geometry::intersection(tile_unclipped, rotated_polygon, boost_tiles))
                continue;

           for (BoostPolygon t : boost_tiles)
           {
                if (bg::area(t) > minTileArea){
                    // Transform boost_tile to world coordinate system.
                    BoostPolygon rotated_tile;
                    BoostPolygon translated_tile;
                    boost::geometry::transform(t, rotated_tile, rotate_back);
                    boost::geometry::transform(rotated_tile, translated_tile, translate_back);

                    // Store tile and calculate center point.
                    tiles.push_back(translated_tile);
                }
           }
        }
    }

    if (tiles.size() < 1){
        errorString = "No tiles calculated. Is the minTileArea parameter large enough?";
        return false;
    }

    return true;
}

bool joinAreas(const std::vector<BoostPolygon> &areas, BoostPolygon &joinedArea)
{
    if (areas.size() < 1)
        return false;
    std::deque<std::size_t> idxList;
    for(size_t i = 1; i < areas.size(); ++i)
        idxList.push_back(i);
    BoostPolygon partialArea = areas[0];

    std::deque<BoostPolygon> sol;
    while (idxList.size() > 0){
        bool success = false;
        for (auto it = idxList.begin(); it != idxList.end(); ++it){
            bg::union_(partialArea, areas[*it], sol);
            if (sol.size() > 0) {
                partialArea = sol[0];
                sol.clear;
                idxList.erase(it);
                success = true;
                break;
            }
        }
        if ( !success )
            return false;
    }
}

bool waypoints(const BoostPolygon &mArea,
               const BoostPolygon &jArea,
               std::vector<BoostPolygon> &tiles,
               std::vector<long> &progress,
               BoostPoint &home,
               double lineDistance,
               double minTransectLength,
               std::vector<BoostPoint>,
               size_t arrivalPathLength,
               size_t returnPathLength)
{
#ifdef SHOW_TIME
    auto start = std::chrono::high_resolution_clock::now();
#endif
    if (!_generateTransects(mArea, lineDistance, minTransectLength, transects))
        return false;
#ifdef SHOW_TIME
    auto delta = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::high_resolution_clock::now() - start);
    cout << endl;
   cout << "Execution time _generateTransects(): " << delta.count() << " ms" << endl;
#endif

    //=======================================
    // Route Transects using Google or-tools.
    //=======================================
    // Offset joined area.
    BoostPolygon jAreaOffset;
    offsetPolygon(jArea, jAreaOffset, detail::offsetConstant);

    // Create vertex list;
    BoostLineString vertices;
    size_t n_t = transects.size()*2;
    size_t n0  = n_t+1;
    vertices.reserve(n0);
    for (auto lstring : transects){
        for (auto vertex : lstring){
            vertices.push_back(vertex);
        }
    }
    vertices.push_back(home);

    for (long i=0; i<long(jArea.outer().size())-1; ++i) {
        vertices.push_back(jArea.outer()[i]);
    }
    for (auto ring : jArea.inners()) {
        for (auto vertex : ring)
            vertices.push_back(vertex);
    }
    size_t n1 = vertices.size();
    // Generate routing model.
    RoutingDataModel dataModel;
    Matrix<double> connectionGraph(n1, n1);

#ifdef SHOW_TIME
    start = std::chrono::high_resolution_clock::now();
#endif
    _generateRoutingModel(vertices, jAreaOffset, n0, dataModel, connectionGraph);
#ifdef SHOW_TIME
    delta = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::high_resolution_clock::now() - start);
    cout << "Execution time _generateRoutingModel(): " << delta.count() << " ms" << endl;
#endif

    // Create Routing Index Manager.
    RoutingIndexManager manager(dataModel.distanceMatrix.getN(),
                                dataModel.numVehicles,
                                dataModel.depot);
    // Create Routing Model.
    RoutingModel routing(manager);

    // Create and register a transit callback.
    const int transit_callback_index = routing.RegisterTransitCallback(
        [&dataModel, &manager](int64 from_index, int64 to_index) -> int64 {
          // Convert from routing variable Index to distance matrix NodeIndex.
          auto from_node = manager.IndexToNode(from_index).value();
          auto to_node = manager.IndexToNode(to_index).value();
          return dataModel.distanceMatrix.get(from_node, to_node);
        });

    // Define cost of each arc.
    routing.SetArcCostEvaluatorOfAllVehicles(transit_callback_index);


    // Define Constraints.
    size_t n = _transects.size()*2;
    Solver *solver = routing.solver();
    for (size_t i=0; i<n; i=i+2){
//        auto idx0 = manager.NodeToIndex(RoutingIndexManager::NodeIndex(i));
//        auto idx1 = manager.NodeToIndex(RoutingIndexManager::NodeIndex(i+1));
//        auto cond0 = routing.NextVar(idx0)->IsEqual(idx1);
//        auto cond1 = routing.NextVar(idx1)->IsEqual(idx0);
//        auto c = solver->MakeNonEquality(cond0, cond1);
//        solver->AddConstraint(c);

        // alternative
        auto idx0 = manager.NodeToIndex(RoutingIndexManager::NodeIndex(i));
        auto idx1 = manager.NodeToIndex(RoutingIndexManager::NodeIndex(i+1));
        auto cond0 = routing.NextVar(idx0)->IsEqual(idx1);
        auto cond1 = routing.NextVar(idx1)->IsEqual(idx0);
        vector<IntVar*> conds{cond0, cond1};
        auto c = solver->MakeAllDifferent(conds);
        solver->MakeRejectFilter();
        solver->AddConstraint(c);
    }


    // Setting first solution heuristic.
    RoutingSearchParameters searchParameters = DefaultRoutingSearchParameters();
    searchParameters.set_first_solution_strategy(
        FirstSolutionStrategy::PATH_CHEAPEST_ARC);
    google::protobuf::Duration *tMax = new google::protobuf::Duration(); // seconds
    tMax->set_seconds(10);
    searchParameters.set_allocated_time_limit(tMax);

    // Solve the problem.
#ifdef SHOW_TIME
    start = std::chrono::high_resolution_clock::now();
#endif
    const Assignment* solution = routing.SolveWithParameters(searchParameters);
#ifdef SHOW_TIME
    delta = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::high_resolution_clock::now() - start);
    cout << "Execution time routing.SolveWithParameters(): " << delta.count() << " ms" << endl;
#endif

    if (!solution || solution->Size() <= 1){
        errorString = "Not able to solve the routing problem.";
        return false;
    }

    // Extract waypoints from solution.
    long index = routing.Start(0);
    std::vector<size_t> route;
    route.push_back(manager.IndexToNode(index).value());
    while (!routing.IsEnd(index)){
        index = solution->Value(routing.NextVar(index));
        route.push_back(manager.IndexToNode(index).value());
    }

    // Connect transects
    waypoint.push_back(vertices[route[0]]);
    vector<size_t> pathIdx;
    arrivalPathLength = 0;
    for (long i=0; i<long(route.size())-1; ++i){
        size_t idx0 = route[i];
        size_t idx1 = route[i+1];
        pathIdx.clear();
        shortestPathFromGraph(connectionGraph, idx0, idx1, pathIdx);
        if ( i==0 )
            arrivalPathLength = pathIdx.size();
        for (size_t j=1; j<pathIdx.size(); ++j)
            waypoints.push_back(vertices[pathIdx[j]]);
    }

    returnPathLength = pathIdx.size();
    return true;

}

bool FlightPlan::_generateTransects(double lineDistance,
                                    double minTransectLength,
                                    const BoostPolygon mArea,
                                    const std::vector<BoostPolygon> &tiles,
                                    const std::vector<long> &progress,
                                    vector<BoostLineString> &transects,
                                    std::string &errorString)
{
    if (tiles.size() != progress.size()){
        ostringstream strstream;
        strstream << "Number of tiles ("
                  << _scenario.getTilesENU().size()
                  << ") is not equal to progress array length ("
                  << _progress.size()
                  << ")";
        errorString = strstream.str();
        return false;
    }

    // Calculate processed tiles (_progress[i] == 100) and subtract them from measurement area.
    size_t num_tiles = progress.size();
    vector<BoostPolygon> processedTiles;
    {
        for (size_t i = 0; i < num_tiles; ++i) {
            if (progress[i] == 100){
                processedTiles.push_back(tiles[i]);
            }
        }

        if (processedTiles.size() == num_tiles)
            return true;
    }

    // Convert measurement area and tiles to clipper path.
    ClipperLib::Path mAreaClipper;
    for ( auto vertex : mArea.outer() ){
        mAreaClipper.push_back(
                    ClipperLib::IntPoint{
                        static_cast<ClipperLib::cInt>(vertex.get<0>()*CLIPPER_SCALE),
                        static_cast<ClipperLib::cInt>(vertex.get<1>()*CLIPPER_SCALE)
                    }
                    );
    }
    vector<ClipperLib::Path> processedTilesClipper;
    for (auto t : processedTiles){
        ClipperLib::Path path;
        for (auto vertex : t.outer()){
            path.push_back(ClipperLib::IntPoint{static_cast<ClipperLib::cInt>(vertex.get<0>()*CLIPPER_SCALE),
                                                static_cast<ClipperLib::cInt>(vertex.get<1>()*CLIPPER_SCALE)});
        }
        processedTilesClipper.push_back(path);
    }

    const BoundingBox bbox;
    minimalBoundingBox(mArea, bbox);
    double alpha                = bbox.angle;
    double x0                   = bbox.corners.outer()[0].get<0>();
    double y0                   = bbox.corners.outer()[0].get<1>();
    double bboxWidth            = bbox.width;
    double bboxHeight           = bbox.height;
    double delta                = detail::offsetConstant;

    size_t num_t = int(ceil((bboxHeight + 2*delta)/lineDistance)); // number of transects
    vector<double> yCoords;
    yCoords.reserve(num_t);
    double y = -delta;
    for (size_t i=0; i < num_t; ++i) {
        yCoords.push_back(y);
        y += lineDistance;
    }


    // Generate transects and convert them to clipper path.
    trans::rotate_transformer<boost::geometry::degree, double, 2, 2> rotate_back(-alpha*180/M_PI);
    trans::translate_transformer<double, 2, 2> translate_back(x0, y0);
    vector<ClipperLib::Path> transectsClipper;
    transectsClipper.reserve(num_t);
    for (size_t i=0; i < num_t; ++i) {
        // calculate transect
        BoostPoint v1{-delta, yCoords[i]};
        BoostPoint v2{bboxWidth+delta, yCoords[i]};
        BoostLineString transect;
        transect.push_back(v1);
        transect.push_back(v2);

        // transform back
        BoostLineString temp_transect;
        bg::transform(transect, temp_transect, rotate_back);
        transect.clear();
        bg::transform(temp_transect, transect, translate_back);


        ClipperLib::IntPoint c1{static_cast<ClipperLib::cInt>(transect[0].get<0>()*CLIPPER_SCALE),
                                static_cast<ClipperLib::cInt>(transect[0].get<1>()*CLIPPER_SCALE)};
        ClipperLib::IntPoint c2{static_cast<ClipperLib::cInt>(transect[1].get<0>()*CLIPPER_SCALE),
                                static_cast<ClipperLib::cInt>(transect[1].get<1>()*CLIPPER_SCALE)};
        ClipperLib::Path path{c1, c2};
        transectsClipper.push_back(path);
    }

    // Perform clipping.
    // Clip transects to measurement area.
    ClipperLib::Clipper clipper;
    clipper.AddPath(mAreaClipper, ClipperLib::ptClip, true);
    clipper.AddPaths(transectsClipper, ClipperLib::ptSubject, false);
    ClipperLib::PolyTree clippedTransecsPolyTree1;
    clipper.Execute(ClipperLib::ctIntersection, clippedTransecsPolyTree1, ClipperLib::pftNonZero, ClipperLib::pftNonZero);

    // Subtract holes (tiles with measurement_progress == 100) from transects.
    clipper.Clear();
    for (auto child : clippedTransecsPolyTree1.Childs)
        clipper.AddPath(child->Contour, ClipperLib::ptSubject, false);
    clipper.AddPaths(processedTilesClipper, ClipperLib::ptClip, true);
    ClipperLib::PolyTree clippedTransecsPolyTree2;
    clipper.Execute(ClipperLib::ctDifference, clippedTransecsPolyTree2, ClipperLib::pftNonZero, ClipperLib::pftNonZero);

    // Extract transects from  PolyTree and convert them to BoostLineString
    for (auto child : clippedTransecsPolyTree2.Childs){
        ClipperLib::Path clipperTransect = child->Contour;
        BoostPoint v1{static_cast<double>(clipperTransect[0].X)/CLIPPER_SCALE,
                      static_cast<double>(clipperTransect[0].Y)/CLIPPER_SCALE};
        BoostPoint v2{static_cast<double>(clipperTransect[1].X)/CLIPPER_SCALE,
                      static_cast<double>(clipperTransect[1].Y)/CLIPPER_SCALE};

        BoostLineString transect{v1, v2};
        if (bg::length(transect) >= minTransectLength)
            transects.push_back(transect);

    }

    if (transects.size() < 1)
        return false;

    return true;
}

}
